@@ -3,10 +3,14 @@ use strict;
 use warnings;
 
 use Config;
+use constant OSNAME => $^O;
 
 sub link_files {
   my $shared_dir = shift;
-  my $exlude_tests = shift;
+  my $do_tests = shift || "";
+  my $exclude_tests= $do_tests eq "without_tests";
+  my $tests_only= $do_tests eq "tests_only";
+
   # This fires from a git source tree only.
   # Right now, all devs are on Linux. Feel free to make portable.
   eval {
@@ -20,7 +24,8 @@ sub link_files {
           my $f = $_;
           s/^\Q$shared_dir\E\/?// or die $_;
           return unless $_;
-          return if $exlude_tests && m#^/?t/#;
+          return if $exclude_tests && m#^/?t/#;
+          return if $tests_only && !m#^/?t/#;
 
           if (-d $f) {
             File::Path::mkpath($_)
@@ -32,7 +37,14 @@ sub link_files {
             my $ref = join "/", ("..") x scalar(@d);
             my $subd = join "/", @d;
             chdir $subd if length($ref);
-            symlink(join("/", grep length, $ref, $shared_dir, $subd, $fname), $fname);
+            my $srcfname = join("/", grep length, $ref, $shared_dir, $subd, $fname);
+            if (OSNAME eq 'MSWin32') {
+              die "link($srcfname, $fname) failed: $!"
+                unless link($srcfname, $fname); #only NTFS implements it
+            }
+            else {
+              symlink($srcfname, $fname);
+            }
             chdir($ref) if length($ref);
           }
         },
