@@ -534,8 +534,15 @@ srl_build_encoder_struct(pTHX_ HV *opt, sv_with_hash *options)
         my_hv_fetchs(he, val, opt, SRL_ENC_OPT_IDX_SORT_KEYS);
         if ( !val )
             my_hv_fetchs(he, val, opt, SRL_ENC_OPT_IDX_CANONICAL);
-        if ( val && SvTRUE(val) )
+        if ( val && SvTRUE(val) ) {
             SRL_ENC_SET_OPTION(enc, SRL_F_SORT_KEYS);
+            if (SvIV(val) > 1) {
+                SRL_ENC_SET_OPTION(enc, SRL_F_SORT_KEYS_PERL);
+                if (SvIV(val) > 2) {
+                    SRL_ENC_SET_OPTION(enc, SRL_F_SORT_KEYS_PERL_REV);
+                }
+            }
+        }
 
         my_hv_fetchs(he, val, opt, SRL_ENC_OPT_IDX_CANONICAL_REFS);
         if ( !val )
@@ -1157,24 +1164,6 @@ srl_dump_av(pTHX_ srl_encoder_t *enc, AV *src, U32 refcount)
     }
 }
 
-#define SIGN(cmp) (cmp < 0 ? -1 : cmp == 0 ? 0 : 1)
-
-SRL_STATIC_INLINE int
-he_cmp_fast_inline(HE *a, HE *b)
-{
-    /* no need for a dTHX here, we don't use anything that needs it */
-    STRLEN la = HeKLEN(a);
-    STRLEN lb = HeKLEN(b);
-
-    int cmp = memcmp(HeKEY(a), HeKEY(b), la < lb ? la : lb);
-    if (!cmp)
-        cmp = la - lb; /* "a" should less than "aa", so 1 - 2 == -1 */
-    return cmp;
-}
-#define ISLT_HE_FAST(a,b)   ( he_cmp_fast_inline( *a, *b ) < 0 )
-#define ISLT_HE_SLOW(a,b)   ( sv_cmp( HeSVKEY_force( *a ), HeSVKEY_force( *b ) ) < 0 )
-#define ISLT_KV(a,b)        ( sv_cmp( a->key, b->key ) < 0 )
-
 SRL_STATIC_INLINE void
 srl_dump_hv_unsorted_nomg(pTHX_ srl_encoder_t *enc, HV *src, UV n)
 {
@@ -1219,58 +1208,153 @@ srl_dump_hv_unsorted_mg(pTHX_ srl_encoder_t *enc, HV *src, const UV n)
         croak("Panic: cannot serialize a tied hash which changes its size!");
 }
 
-#define CALL_QSORT_NOBYTES(TYPE, ARY, N, F)                 \
-    STMT_START {                                            \
-        /* hack to forcefully disable "use bytes" */        \
-        COP cop= *PL_curcop;                                \
-        cop.op_private= 0;                                  \
-                                                            \
-        ENTER;                                              \
-        SAVETMPS;                                           \
-                                                            \
-        SAVEVPTR (PL_curcop);                               \
-        PL_curcop= &cop;                                    \
-                                                            \
-        QSORT(TYPE, ARY, N, F)                              \
-                                                            \
-        FREETMPS;                                           \
-        LEAVE;                                              \
-    } STMT_END
+/* sorting hashes - nothing in perl is easy. ever.
+ *
+ * Some things to keep in mind about perl hashes as you read this code:
+ *
+ * Hashes may be shared or not. Usually shared. This means they share their
+ * key data via PL_strtab.
+ *
+ * Hashes may be tied or not. Usually not. When tied the keys from the hash
+ * are available only as SV *'s, and when untied, the keys from the hash are
+ * accessed via HE *'s.
+ *
+ * Some HE's actually contains SV's but most contain a ptr/len combo with
+ * an utf8 flag. To make things even more interesting utf8 keys are
+ * normalized to latin1 by perl where possible before being stored in the HE,
+ * with the utf8 flag indicating "was utf8" instead of "is utf8" or "not utf8".
+ *
+ * The complexity about accessing the key for a hash can be managed away by
+ * perl via API's like hv_iterkeysv(), but using that means constructing mortal
+ * SV's for each key as we go.
+ *
+ * We could in theory use the HePV() interface, but one annoying result of the
+ * "was utf8" logic is that we can't use a sort comparator which looks
+ * at the raw binary of the keys when the keys might contain utf8. A utf8 key
+ * like "\xDF" will be downgraded to ascii in the HE form, but will be upgraded
+ * to the utf8 representation in the SV form. So if we want to do "fast" sorting
+ * we have to restrict it to non-utf8/non-sv keys, and force the use of the SV
+ * based API (which we have to use for tie's anyway) when we see a UTF8 key.
+ *
+ * Which is what we do below. In order to sort a hash we need to construct an
+ * array of its contents, in srl_dump_sorted_nomg() we walk the hash, checking
+ * each key, and copying each HE over into a scratch buffer which it then sorts.
+ * If during the transcription process it sees any utf8 or SV keys it exits
+ * immediately, and falls through to srl_dump_sort_mg(), which uses hv_iterkeysv()
+ * to construct an array of HE_SV instead, which we then sort.
+ */
+
+
+
+SRL_STATIC_INLINE int
+he_islt(const HE *a, const HE *b)
+{
+    /* no need for a dTHX here, we don't use anything that needs it */
+    const STRLEN la = HeKLEN(a);
+    const STRLEN lb = HeKLEN(b);
+    const int cmp = memcmp(HeKEY(a), HeKEY(b), la < lb ? la : lb);
+    if (cmp) {
+        return cmp < 0;
+    } else {
+        return la < lb;
+    }
+}
+
+SRL_STATIC_INLINE int
+he_sv_islt_fast(const HE_SV *a, const HE_SV *b)
+{
+    /* no need for a dTHX here, we don't use anything that needs it */
+    char *a_ptr;
+    char *b_ptr;
+    int a_isutf8;
+    int b_isutf8;
+    const STRLEN a_len= a->key.sv ? SvCUR(a->key.sv) : HeKLEN(a->val.he);
+    const STRLEN b_len= b->key.sv ? SvCUR(b->key.sv) : HeKLEN(b->val.he);
+    if (a_len != b_len) {
+        return a_len < b_len;
+    }
+    a_isutf8= (a->key.sv ? SvUTF8(a->key.sv) : HeKUTF8(a->val.he)) ? 0 : 1;
+    b_isutf8= (b->key.sv ? SvUTF8(b->key.sv) : HeKUTF8(b->val.he)) ? 0 : 1;
+    if (a_isutf8 != b_isutf8) {
+        return a_isutf8 < b_isutf8;
+    }
+    a_ptr= a->key.sv ? SvPVX(a->key.sv) : HeKEY(a->val.he);
+    b_ptr= b->key.sv ? SvPVX(b->key.sv) : HeKEY(b->val.he);
+    return memcmp(a_ptr, b_ptr, a_len < b_len ? a_len : b_len ) < 0;
+}
+
+#define ISLT_HE_SV(a,b)    he_sv_islt_fast( a, b )
+#define ISLT_SV_CMP(a,b)   sv_cmp(a->key.sv, b->key.sv) == sort_dir
+
 
 SRL_STATIC_INLINE void
-srl_dump_hv_sorted_mg(pTHX_ srl_encoder_t *enc, HV *src, const UV n)
+srl_qsort(pTHX_ srl_encoder_t *enc, const UV n, HE_SV *array)
+{
+    if ( SRL_ENC_HAVE_OPTION(enc, SRL_F_SORT_KEYS_PERL) ) {
+        int sort_dir= SRL_ENC_HAVE_OPTION(enc, SRL_F_SORT_KEYS_PERL_REV) ? 1 : -1;
+        /* hack to forcefully disable "use bytes" */
+        COP cop= *PL_curcop;
+        cop.op_private= 0;
+
+        ENTER;
+        SAVETMPS;
+
+        SAVEVPTR (PL_curcop);
+        PL_curcop= &cop;
+       
+        /* now sort */
+        QSORT(HE_SV, array, n, ISLT_SV_CMP);
+
+        FREETMPS;
+        LEAVE;
+    } else {
+        /* now sort */
+        QSORT(HE_SV, array, n, ISLT_HE_SV);
+    }
+}
+
+
+SRL_STATIC_INLINE void
+srl_dump_hv_sorted_sv_slow(pTHX_ srl_encoder_t *enc, HV *src, const UV n, HE_SV *array)
 {
     HE *he;
     UV i= 0;
     const int do_share_keys = HvSHAREKEYS((SV *)src);
+    const int is_tie= !array;
+
+    /* This sub is used for ties, and for hashes with SV keys in them,
+     * and when the user requests SORT_KEYS_PERL, it is the slowest way
+     * and most memory hungry way to serialize a hash. We will use the 
+     * full perl api for extracting the contents of the hash, which fortifies
+     * us against ties, and we will convert all keys into mortal
+     * sv's where necessary. This means we can use sv_cmp on the keys
+     * if we wish.
+     */
 
     (void)hv_iterinit(src); /* return value not reliable according to API docs */
-    /* can't use the same optimizations we use for normal hashes with tied
-     * hashes - the HE*'s that are returned for tied hashes are temporary
-     * structures returned from the tie infra. So we make an array, and
-     * then sort it.*/
     {
-        kv_sv *kv_sv_array;
-        kv_sv *kv_sv_array_end;
-        Newx(kv_sv_array, n, kv_sv);
-        SAVEFREEPV(kv_sv_array);
-        kv_sv_array_end= kv_sv_array + n;
+        HE_SV *array_end;
+        if (!array) {
+            Newx(array, n, HE_SV);
+            SAVEFREEPV(array);
+        }
+        array_end= array + n;
         while ((he = hv_iternext(src))) {
             if (expect_false( i == n ))
-                croak("Panic: cannot serialize a tied hash which changes its size!");
-            kv_sv_array[i].key= hv_iterkeysv(he);
-            kv_sv_array[i].val= hv_iterval(src,he);
+                croak("Panic: cannot serialize a %s hash which changes its size!",is_tie ? "tied" : "untied");
+            array[i].key.sv= hv_iterkeysv(he);
+            array[i].val.sv= hv_iterval(src,he);
             i++;
         }
         if (expect_false( i != n ))
-            croak("Panic: can not serialize a tied hash which changes it size!");
+            croak("Panic: can not serialize a %s hash which changes it size!", is_tie ? "tied" : "untied");
 
-        CALL_QSORT_NOBYTES(kv_sv, kv_sv_array, n, ISLT_KV);
+        srl_qsort(aTHX_ enc, n, array);
 
-        while ( kv_sv_array < kv_sv_array_end ) {
-            CALL_SRL_DUMP_SV(enc, kv_sv_array->key);
-            CALL_SRL_DUMP_SV(enc, kv_sv_array->val);
-            kv_sv_array++;
+        while ( array < array_end ) {
+            CALL_SRL_DUMP_SV(enc, array->key.sv);
+            CALL_SRL_DUMP_SV(enc, array->val.sv);
+            array++;
         }
     }
 }
@@ -1282,29 +1366,34 @@ srl_dump_hv_sorted_nomg(pTHX_ srl_encoder_t *enc, HV *src, const UV n)
     HE *he;
     const int do_share_keys = HvSHAREKEYS((SV *)src);
 
+    /* This sub is used only for untied hashes and when the user wants
+     * sorted keys, but not necessarily the order that perl would use. 
+     */
+
     (void)hv_iterinit(src); /* return value not reliable according to API docs */
     {
-        HE **he_array;
-        HE **he_array_ptr;
-        HE **he_array_end;
-        int fast = 1;
-        Newxz(he_array, n, HE*);
-        SAVEFREEPV(he_array);
-        he_array_ptr = he_array;
-        he_array_end = he_array + n;
+        HE_SV *array;
+        HE_SV *array_ptr;
+        HE_SV *array_end;
+        Newx(array, n, HE_SV);
+        SAVEFREEPV(array);
+        array_ptr = array;
         while ((he = hv_iternext(src))) {
-            *he_array_ptr++ = he;
-            if ( HeKLEN(he) < 0 || HeKUTF8(he) )
-                fast = 0;
+            if ( HeKWASUTF8(he) ) {
+                array_ptr->key.sv= hv_iterkeysv(he);
+            } else {
+                array_ptr->key.sv = HeSVKEY(he);
+            }
+            array_ptr->val.he = he;
+            array_ptr++;
         }
-        if (fast) {
-            QSORT(HE *, he_array, n, ISLT_HE_FAST)
-        } else {
-            CALL_QSORT_NOBYTES(HE *, he_array, n, ISLT_HE_SLOW);
-        }
-        while ( he_array < he_array_end ) {
+        
+        srl_qsort(aTHX_ enc, n, array);
+
+        array_end = array + n;
+        for ( array_end= array + n; array < array_end; array++ ) {
             SV *v;
-            he = *he_array++;
+            he = array->val.he;
             v = hv_iterval(src, he);
             srl_dump_hk(aTHX_ enc, he, do_share_keys);
             CALL_SRL_DUMP_SV(enc, v);
@@ -1336,9 +1425,11 @@ srl_dump_hv(pTHX_ srl_encoder_t *enc, HV *src, U32 refcount)
     }
 
     if ( n ) {
-        if ( SvMAGICAL(src) ) {
-            if ( SRL_ENC_HAVE_OPTION(enc, SRL_F_SORT_KEYS) ) {
-                srl_dump_hv_sorted_mg(aTHX_ enc, src, n);
+        if ( SvMAGICAL(src) || SRL_ENC_HAVE_OPTION(enc, SRL_F_SORT_KEYS_PERL) ) {
+            /* SORT_KEYS_PERL implies SORT_KEYS, but we check for either just to be
+             * careful - yves*/
+            if ( SRL_ENC_HAVE_OPTION(enc, SRL_F_SORT_KEYS|SRL_F_SORT_KEYS_PERL) ) {
+                srl_dump_hv_sorted_sv_slow(aTHX_ enc, src, n, NULL);
             }
             else {
                 srl_dump_hv_unsorted_mg(aTHX_ enc, src, n);
